/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;

import static java.util.Objects.requireNonNull;

/**
 * <p>Capacity bounded array of objects.</p>
 *
 * <p>Bounded object arrays are constructed by the static interface methods of this interface. The {@code newInstance}
 * method returns a new empty array. The {@code copyOf} method returns a new array representing a shallow copy of the
 * array argument. The {@code valueOf} methods wrap a new {@code BoundedObjectArray} around a native array.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Bounded arrays, such as {@code BoundedByteArray}, always report {@code BOUNDED} and {@code MUTABLE}.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * Characteristics are always invalidated if conditions of these characteristics are broken as result of array
 * modification. Characteristics are only added in scenarios in which their conditions can be 'tested' without
 * traversal of elements. For example, the {@code SORTED} characteristic is added if the length of an array is reduced
 * to zero or one element, or if all elements are replaced by an array that is marked as being sorted.
 * The semantics of characteristics processing is documented for each "destructive" method. All implementations returned
 * by the framework adhere to these semantics. Custom implementations may choose to set characteristics at their
 * discretion as long as they do not violate the contract of these characteristics.
 * Users can take full control of characteristics assignment by means of the {@link MutableArray#doAction(ArrayAction)}
 * method.</p>
 *
 * @param <E> element type.
 * @author Leon van Zantvoort
 */
public interface BoundedObjectArray<E> extends DynamicObjectArray<E>, BoundedArray<E> {

    /**
     * Creates an empty array with the specified capacity. The backing array will be of type Object.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code BOUNDED}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param <E> element type.
     * @param capacity the capacity of the array.
     * @throws IllegalArgumentException if the specified capacity is negative.
     * @return a new bounded array.
     */
    @SuppressWarnings("unchecked")
    static <E> BoundedObjectArray<E> newInstance(int capacity) {
        return (BoundedObjectArray<E>) newInstance(Object.class, capacity);
    }

    /**
     * Creates an empty array with the specified capacity. The backing array will be of type {@code cls}.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE}, {@code BOUNDED}, {@code SORTED} and {@code DISTINCT}.</li>
     * </ul>
     *
     * @param <E> element type.
     * @param capacity the capacity of the array.
     * @param cls component type.
     * @throws IllegalArgumentException if the specified capacity is negative.
     * @return a new bounded array.
     * @throws NullPointerException if the specified class is null.
     */
    @SuppressWarnings("unchecked")
    static <E> BoundedObjectArray<E> newInstance(@Nonnull Class<E> cls, int capacity) {
        if (capacity < 0) {
            throw new IllegalArgumentException("Illegal capacity: " + capacity);
        }
        return valueOf((E[]) java.lang.reflect.Array.newInstance(requireNonNull(cls), capacity), 0, 0);
    }

    /**
     * Creates an capacity bounded copy of the specified {@code array}, backed by a native array. If the specified
     * {@code array} is mutable, changes to the mutable array are not reflected by this bounded copy.
     *
     * <p>The specified {@code array} can be any array of primitive or object numbers. Reconstructing a copy of this
     * array may involve rounding as implemented by {@link Number#byteValue}. The specified {@code array}
     * must not contain any {@code null} elements.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE} and {@code BOUNDED};</li>
     *   <li>{@code NONNULL} if the specified {@code array} contains zero elements;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}.</li>
     * </ul>
     *
     * @param <E> element type.
     * @param array array to be copied.
     * @return a bounded array of objects.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings({"unchecked", "InconsistentJavaDoc"})
    static <E> BoundedObjectArray<E> copyOf(@Nonnull Array<E> array) {
        E[] backingArray = (E[]) NativeArrays.array(array);
        if (backingArray != null) {
            return Arrays.newBoundedArray(backingArray, array.offset(), array.length(), array.characteristics(),
                    array.getComparator(), true);
        }
        return Arrays.newBoundedArray((E[]) NativeArrays.toArray(array), 0, array.length(), array.characteristics(),
                array.getComparator(), false);
    }

    /**
     * Creates an bounded array of the specified {@code array}. The specified {@code array} is used as backing array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE} and {@code BOUNDED};</li>
     *   <li>{@code NONNULL} if the specified {@code array} contains zero elements;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param <E> element type.
     * @param array array to be wrapped.
     * @return an bounded array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @SafeVarargs
    static <E> BoundedObjectArray<E> valueOf(@Nonnull E... array) {
        return valueOf(array, 0, array.length);
    }

    /**
     * Creates an bounded array of the specified {@code array}. The specified {@code array} is used as backing array.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code NONNULL}, {@code MUTABLE} and {@code BOUNDED};</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param <E> element type.
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @return an bounded array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    static <E> BoundedObjectArray<E> valueOf(@Nonnull E[] array, int offset, int length) {
        return Arrays.newBoundedArray(array, offset, length, 0, null, false);
    }

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    BoundedObjectArray<E> offset(int offset);     // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    MutableObjectArray<E> length(int length);     // Should not return bounded array!

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    MutableObjectArray<E> subArray(int fromIndex, int toIndex);   // Should not return bounded array!
}
