/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.Nonnull;
import java.util.Comparator;

/**
 * <p>Fixed-size mutable array of object references.</p>
 *
 * <p>Mutable object arrays are constructed by the static interface methods of this interface. The {@code copyOf}
 * method returns a new array representing a shallow copy of the array argument. The {@code valueOf} methods wrap a new
 * {@code MutableObjectArray} around a native array.</p>
 *
 * <p>Arrays express characteristics to indicate that, for instance, elements are listed in sorted order, or no
 * duplicate elements exist. These characteristics allow array implementations to optimize their algorithms based
 * on the contents of the array. For example, operations that lookup elements can use a binary search algorithm if
 * the array is marked as being sorted. These optimizations may reduce the time complexity of such operations by
 * several orders of magnitude.
 * Mutable arrays, such as {@code MutableObjectArray}, always report {@code MUTABLE}.</p>
 *
 * <p>The array implementations of the framework attempt to establish these characteristics to accommodate these
 * optimizations at best effort. This means that the array implementation determines which set of characteristics apply
 * to its elements without traversing these elements, in other words, without incurring a performance penalty.
 * Characteristics are always invalidated if conditions of these characteristics are broken as result of array
 * modification. Characteristics are only added in scenarios in which their conditions can be 'tested' without
 * traversal of elements. For example, the {@code SORTED} characteristic is added if the length of an array is reduced
 * to zero or one element, or if all elements are replaced by an array that is marked as being sorted.
 * The semantics of characteristics processing is documented for each "destructive" method. All implementations returned
 * by the framework adhere to these semantics. Custom implementations may choose to set characteristics at their
 * discretion as long as they do not violate the contract of these characteristics.
 * Users can take full control of characteristics assignment by means of the {@link MutableArray#doAction(ArrayAction)}
 * method.</p>
 *
 * @param <E> element type.
 * @author Leon van Zantvoort
 */
public interface MutableObjectArray<E> extends ObjectArray<E>, MutableArray<E> {

    /**
     * Creates a fixed-size mutable copy of the specified {@code array}, backed by a native array.
     * If the specified {@code array} is mutable, changes to the mutable array are not reflected by this mutable copy.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE};</li>
     *   <li>{@code NONNULL} if the specified {@code array} contains zero elements;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s);</li>
     *   <li>all characteristics reported by the specified {@code array}, except {@code BOUNDED}.</li>
     * </ul>
     *
     * @param <E> element type.
     * @param array array to be copied.
     * @return fixed-size mutable array of objects.
     * @throws NullPointerException if the specified array is null.
     */
    @SuppressWarnings({"unchecked", "InconsistentJavaDoc"})
    static <E> MutableObjectArray<E> copyOf(@Nonnull Array<E> array) {
        E[] backingArray = (E[]) NativeArrays.array(array);
        if (backingArray != null) {
            return Arrays.newMutableArray(backingArray, array.offset(), array.length(), array.characteristics(),
                    array.getComparator(), true);
        }
        return Arrays.newMutableArray((E[]) NativeArrays.toArray(array), 0, array.length(), array.characteristics(),
                array.getComparator(), false);
    }

    /**
     * Creates a mutable array of the specified {@code array}. The specified {@code array} is used as backing array.
     *
     * <p>The contents of {@code array} must not be modified after this method returns as this would violate this
     * array's characteristics contract.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE};</li>
     *   <li>{@code NONNULL} if the specified {@code array} contains zero elements;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code array} contains zero or one element(s).</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param <E> element type.
     * @return a mutable array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null.
     */
    @SafeVarargs
    static <E> MutableObjectArray<E> valueOf(@Nonnull E... array) {
        return valueOf(array, 0, array.length);
    }

    /**
     * Creates a mutable array of the specified {@code array}. The specified {@code array} is used as backing array.
     *
     * <p>The contents of {@code array} must not be modified after this method returns as this would violate this
     * array's characteristics contract.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>The newly constructed array reports the following characteristics:</p>
     *
     * <ul>
     *   <li>{@code MUTABLE};</li>
     *   <li>{@code NONNULL} if the specified {@code length} is zero;</li>
     *   <li>{@code SORTED} and {@code DISTINCT} if the specified {@code length} is zero or one.</li>
     * </ul>
     *
     * @param array array to be wrapped.
     * @param offset specifies the first element of this array's range.
     * @param length specifies the range length of this array.
     * @param <E> element type.
     * @return a mutable array that is backed by the specified {@code array}.
     * @throws NullPointerException if the specified array is null or if the array contains null elements and this
     * array does not permit null elements.
     * @throws IndexOutOfBoundsException if offset or length is invalid.
     */
    static <E> MutableObjectArray<E> valueOf(@Nonnull E[] array, int offset, int length) {
        return Arrays.newMutableArray(array, offset, length, 0, null, false);
    }

    /**
     * Sorts this array using the supplied {@code comparator} to compare elements. The {@code comparator} can be
     * retrieved by {@code getComparator} after this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     *
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> should be used.
     * @throws IllegalArgumentException (optional) if the comparator is found to violate the
     * {@link java.util.Comparator} contract.
     * @throws NullPointerException if this array contains null elements and comparator does not permit null arguments.
     * @throws ClassCastException if this array contains non {@code Comparable} elements or comparator not compatible
     * with elements.
     */
    void sort(Comparator<? super E> comparator);

    /**
     * Sorts this array using the supplied {@code Comparator} to compare elements. The underlying implementation
     * utilizes multiple threads to reduce sorting time. The algorithm requires a working space equal to the size of the
     * this array. The {@code comparator} can be retrieved by {@code getComparator} after this call.
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>Modification of this array may affect the array's characteristics. The array implementations of the
     * Java Array Framework, as returned by the static factory methods, respect the characteristics of the array after
     * modification:</p>
     *
     * <ul>
     *   <li>{@code SORTED} is reported after this invocation.</li>
     * </ul>
     *
     * @param comparator the {@code Comparator} used to compare array elements. A {@code null} value indicates that the
     *                   elements' <i>Comparable natural ordering</i> should be used.
     * @throws IllegalArgumentException (optional) if the comparator is found to violate the
     * {@link java.util.Comparator} contract.
     * @throws NullPointerException if this array contains null elements and comparator does not permit null arguments.
     * @throws ClassCastException if this array contains non {@code Comparable} elements or comparator not compatible
     * with elements.
     */
    void parallelSort(Comparator<? super E> comparator);

    /**
     * Returns the backing array (optional operation). All elements of {@code this} mutable array and parents of this
     * mutable array are stored in the returned object. Modifications to the backing array are reflected by {@code this}
     * mutable array and it's parents.
     *
     * <p>The resulting object must be cast to the primitive or object type of the implementing class.</p>
     *
     * <p><strong>Characteristics:</strong></p>
     *
     * <p>This method allows the caller to modify this array outside the scope of the implementing class. Hence, it can
     * not guarantee correctness of the characteristics of this array. For that reason, content related characteristics
     * (such as {@code NONNULL}, {@code SORTED} and {@code DISTINCT}) are removed by calling this method.
     * Characteristics removal can be suppressed by obtaining the backing array through {@code doAction}:</p>
     *
     * <pre>
     *{@literal Object[] backingArray = (Object[]) array.doAction(ArrayContext::array);}</pre>
     *
     * @return backing array.
     * @throws UnsupportedOperationException if array does not expose backing array.
     */
    @SuppressWarnings("InconsistentJavaDoc")
    @Override
    Object[] array();   // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code offset}. The specified {@code offset} is relative to
     * the existing offset, as returned by {@link #offset()}. The length of the resulting array is reduced by
     * the number of elements that are added to the current offset:</p>
     *
     * <pre>int newLength = length() - offset;</pre>
     *
     * <p>Effectively, the virtual {@code toIndex} of the array remains the same.</p>
     *
     * @param offset new relative offset.
     * @return a new sub array object with the new {@code offset}.
     */
    @Override
    MutableObjectArray<E> offset(int offset);   // More specific return type.

    /**
     * <p>Returns a new sub array object for the specified {@code length}. Note that {@code length} must be smaller than
     * or equal to the current length, as returned by {@code #length}.</p>
     *
     * @param length new length, must be smaller than or equal to current length.
     * @return a new sub array object with the new {@code length}.
     */
    @Override
    MutableObjectArray<E> length(int length);   // More specific return type.

    /**
     * Convenience method, combining {@code offset(int)} and {@code length(int)}.
     *
     * @param fromIndex low endpoint (inclusive) of the sub array.
     * @param toIndex high endpoint (exclusive) of the sub array.
     * @return a view of the specified range within this array.
     * @throws IndexOutOfBoundsException for an illegal endpoint index value ({@code fromIndex < 0 || toIndex > length ||
     *         fromIndex > toIndex}).
     */
    @Override
    MutableObjectArray<E> subArray(int fromIndex, int toIndex); // More specific return type.
}
