/*
 * ============================================================================
 * GNU Lesser General Public License
 * ============================================================================
 *
 * Square Brackets - Java Array Framework.
 * Copyright (C) 2013  Beatgrid Media B.V.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
 *
 * Beatgrid Media B.V.
 * Sisalbaan 5-A
 * 2352 AZ Leiderdorp
 * The Netherlands
 *
 * info@squarebrackets.org
 * http://squarebrackets.org
 */
package org.squarebrackets;

import javax.annotation.CheckForNull;
import javax.annotation.Nonnull;
import java.util.Iterator;

/**
 * Arrays implementing this interface provide fast access to a native array containing all elements, including elements
 * of parent arrays. This interface allows utility classes such as {@link org.squarebrackets.Arrays} to perform fast read operations on
 * native arrays. The native arrays returned by this interface must only use for read access. Writing is prohibited.
 *
 * @author Leon van Zantvoort
 */
final class NativeArrays {

    /**
     * Class cannot be instantiated.
     */
    private NativeArrays() {
        throw new AssertionError("No!");
    }

    /**
     * Returns {@code true} if the specified array is backed by a native array of the specified component type,
     * {@code false} otherwise.
     *
     * @param array array to be inspected.
     * @param componentType component type to test for.
     * @return {@code true} if the specified array is backed by a native array of the specified component type,
     * {@code false} otherwise.
     */
    static boolean hasArray(@Nonnull Array<?> array, @Nonnull Class<?> componentType) {
        return array instanceof NativeArray && ((NativeArray<?>) array).hasArray() &&
                componentType.isAssignableFrom(array.getComponentType());
    }

    /**
     * Returns the boolean backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type boolean.
     *
     * @param array array to obtain backing array from.
     * @return boolean backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type boolean.
     */
    static @CheckForNull boolean[] booleanArray(@Nonnull Array<?> array) {
        if (hasArray(array, Boolean.TYPE)) {
            return (boolean[]) ((NativeArray) array).nativeArray();
        }
        return null;
    }

    /**
     * Returns the byte backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type byte.
     *
     * @param array array to obtain backing array from.
     * @return byte backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type byte.
     */
    static @CheckForNull byte[] byteArray(@Nonnull Array<?> array) {
        if (hasArray(array, Byte.TYPE)) {
            return (byte[]) ((NativeArray) array).nativeArray();
        }
        return null;
    }

    /**
     * Returns the char backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type char.
     *
     * @param array array to obtain backing array from.
     * @return char backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type char.
     */
    static @CheckForNull char[] charArray(@Nonnull Array<?> array) {
        if (hasArray(array, Character.TYPE)) {
            return (char[]) ((NativeArray) array).nativeArray();
        }
        return null;
    }

    /**
     * Returns the short backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type short.
     *
     * @param array array to obtain backing array from.
     * @return short backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type short.
     */
    static @CheckForNull short[] shortArray(@Nonnull Array<?> array) {
        if (hasArray(array, Short.TYPE)) {
            return (short[]) ((NativeArray) array).nativeArray();
        }
        return null;
    }

    /**
     * Returns the int backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type int.
     *
     * @param array array to obtain backing array from.
     * @return int backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type int.
     */
    static @CheckForNull int[] intArray(@Nonnull Array<?> array) {
        if (hasArray(array, Integer.TYPE)) {
            return (int[]) ((NativeArray) array).nativeArray();
        }
        return null;
    }

    /**
     * Returns the long backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type long.
     *
     * @param array array to obtain backing array from.
     * @return long backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type long.
     */
    static @CheckForNull long[] longArray(@Nonnull Array<?> array) {
        if (hasArray(array, Long.TYPE)) {
            return (long[]) ((NativeArray) array).nativeArray();
        }
        return null;
    }

    /**
     * Returns the float backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type float.
     *
     * @param array array to obtain backing array from.
     * @return float backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type float.
     */
    static @CheckForNull float[] floatArray(@Nonnull Array<?> array) {
        if (hasArray(array, Long.TYPE)) {
            return (float[]) ((NativeArray) array).nativeArray();
        }
        return null;
    }

    /**
     * Returns the double backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type double.
     *
     * @param array array to obtain backing array from.
     * @return double backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not of type double.
     */
    static @CheckForNull double[] doubleArray(@Nonnull Array<?> array) {
        if (hasArray(array, Double.TYPE)) {
            return (double[]) ((NativeArray) array).nativeArray();
        }
        return null;
    }

    /**
     * Returns the object backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not an object array.
     *
     * @param array array to obtain backing array from.
     * @return object backing array of the specified {@code array}, or {@code null} if the specified array does
     * not expose a backing array, or if the backing array is not an object array.
     */
    static @CheckForNull Object[] array(@Nonnull Array<?> array) {
        if (hasArray(array, Object.class)) {
            return (Object[]) ((NativeArray) array).nativeArray();
        }
        return null;
    }

    /**
     * Returns native boolean array of the specified {@code array}. The returned array may be the backing array of the
     * specified array, or a shallow copy, if the specified array does not expose a backing array.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull boolean[] toBooleanArray(@Nonnull Array<? extends Boolean> array) {
        if (array instanceof BooleanArray) {
            return ((BooleanArray) array).toArray();
        }
        final int length = array.length();
        final boolean[] shallowCopy = new boolean[length];
        final Iterator<? extends Boolean> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            shallowCopy[index] = iterator.next(); // Throws NullPointerException for null elements.
        }
        return shallowCopy;
    }

    /**
     * Returns native byte array of the specified {@code array}. The returned array may be the backing array of the
     * specified array, or a shallow copy, if the specified array does not expose a backing array.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull byte[] toByteArray(@Nonnull Array<? extends Number> array) {
        if (array instanceof ByteArray) {
            return ((ByteArray) array).toArray();
        }
        final int length = array.length();
        final byte[] shallowCopy = new byte[length];
        final Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            shallowCopy[index] = iterator.next().byteValue(); // Throws NullPointerException if element is null.
        }
        return shallowCopy;
    }

    /**
     * Returns native char array of the specified {@code array}. The returned array may be the backing array of the
     * specified array, or a shallow copy, if the specified array does not expose a backing array.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull char[] toCharArray(@Nonnull Array<? extends Character> array) {
        if (array instanceof CharArray) {
            return ((CharArray) array).toArray();
        }
        final int length = array.length();
        final char[] shallowCopy = new char[length];
        final Iterator<? extends Character> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            shallowCopy[index] = iterator.next(); // Throws NullPointerException for null elements.
        }
        return shallowCopy;
    }

    /**
     * Returns native short array of the specified {@code array}. The returned array may be the backing array of the
     * specified array, or a shallow copy, if the specified array does not expose a backing array.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull short[] toShortArray(@Nonnull Array<? extends Number> array) {
        if (array instanceof ShortArray) {
            return ((ShortArray) array).toArray();
        }
        final int length = array.length();
        final short[] shallowCopy = new short[length];
        final Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            shallowCopy[index] = iterator.next().shortValue(); // Throws NullPointerException if element is null.
        }
        return shallowCopy;
    }

    /**
     * Returns native int array of the specified {@code array}. The returned array may be the backing array of the
     * specified array, or a shallow copy, if the specified array does not expose a backing array.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull int[] toIntArray(@Nonnull Array<? extends Number> array) {
        if (array instanceof IntArray) {
            return ((IntArray) array).toArray();
        }
        final int length = array.length();
        final int[] shallowCopy = new int[length];
        final Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            shallowCopy[index] = iterator.next().intValue(); // Throws NullPointerException if element is null.
        }
        return shallowCopy;
    }

    /**
     * Returns native long array of the specified {@code array}. The returned array may be the backing array of the
     * specified array, or a shallow copy, if the specified array does not expose a backing array.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull long[] toLongArray(@Nonnull Array<? extends Number> array) {
        if (array instanceof LongArray) {
            return ((LongArray) array).toArray();
        }
        final int length = array.length();
        final long[] shallowCopy = new long[length];
        final Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            shallowCopy[index] = iterator.next().longValue(); // Throws NullPointerException if element is null.
        }
        return shallowCopy;
    }

    /**
     * Returns native float array of the specified {@code array}. The returned array may be the backing array of the
     * specified array, or a shallow copy, if the specified array does not expose a backing array.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull float[] toFloatArray(@Nonnull Array<? extends Number> array) {
        if (array instanceof FloatArray) {
            return ((FloatArray) array).toArray();
        }
        final int length = array.length();
        final float[] shallowCopy = new float[length];
        final Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            shallowCopy[index] = iterator.next().floatValue(); // Throws NullPointerException if element is null.
        }
        return shallowCopy;
    }

    /**
     * Returns native double array of the specified {@code array}. The returned array may be the backing array of the
     * specified array, or a shallow copy, if the specified array does not expose a backing array.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull double[] toDoubleArray(@Nonnull Array<? extends Number> array) {
        if (array instanceof DoubleArray) {
            return ((DoubleArray) array).toArray();
        }
        final int length = array.length();
        final double[] shallowCopy = new double[length];
        final Iterator<? extends Number> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            shallowCopy[index] = iterator.next().doubleValue(); // Throws NullPointerException if element is null.
        }
        return shallowCopy;
    }

    /**
     * Returns native array of the specified {@code array}. The returned array may be the backing array of the
     * specified array, or a shallow copy, if the specified array does not expose a backing array.
     *
     * @param array array to obtain native array from.
     * @return native array.
     * @throws NullPointerException if array contains null elements.
     */
    static @Nonnull Object[] toArray(@Nonnull Array<?> array) {
        if (array instanceof ObjectArray) {
            return ((ObjectArray) array).toArray();
        }
        final int length = array.length();
        final Object[] shallowCopy = new Object[length];
        final Iterator<?> iterator = array.iterator();
        for (int index = 0; index < length; index++) {
            shallowCopy[index] = iterator.next();
        }
        return shallowCopy;
    }
}
